from zope.interface import Interface, implements
from zope.configuration.fields import GlobalObject
from zope.schema import TextLine
from zope.schema import Bool

from repoze.bfg.zcml import route, utility
from repoze.bfg.configuration import Configurator
from repoze.bfg.threadlocal import get_current_registry
from repoze.bfg.path import caller_package

from repoze.bfg.compat import json
from repoze.bfg.security import has_permission

from webob import Response
from cStringIO import StringIO
import sys, traceback
import inspect
import venusian

# form parameters sent by ExtDirect when using a form-submit
# see http://www.sencha.com/products/js/direct.php
FORM_DATA_KEYS = (
    "extAction",
    "extMethod",
    "extTID",
    "extUpload",
    "extType"
)

# response to a file upload cannot be return as application/json, ExtDirect defines
# a special html response body for this use case where the response data is
# added to a textarea for faster JS-side decoding (since etxtarea text is not a DOM node)
FORM_SUBMIT_RESPONSE_TPL = '<html><body><textarea>%s</textarea></body></html>'

def _mk_cb_key(action_name, method_name):
    return action_name + '#' + method_name

class JsonReprEncoder(json.JSONEncoder):
    """ a convenience wrapper for classes that support json_repr() """
    def default(self, obj):
        jr = getattr(obj, 'json_repr', None)
        if jr is None:
            return super(JsonReprEncoder, self).default(obj)
        return jr()

class IExtdirect(Interface):
    """ marker iface for Extdirect utility """
    pass


class Extdirect(object):
    """
    Handles ExtDirect API respresentation and routing.

    The Extdirect accepts a number of arguments: ``app``,
    ``api_path``, ``router_path``, ``namespace``, ``descriptor``
    and ``expose_exceptions``.

    The ``app`` argument is a python package or module used
    which is used to scan for ``extdirect_method`` decorated
    functions/methods once the API is built.

    The ``api_path`` and ``router_path`` arguments are the
    paths/URIs of repoze.bfg views. ``api_path`` renders
    the ExtDirect API, ``router_path`` is the routing endpoint
    for ExtDirect calls.

    If the ``namespace`` argument is passed it will be used in
    the API as Ext namespace (default is 'Ext.app').

    If the ``descriptor`` argument is passed it's used as ExtDirect
    API descriptor name (default is Ext.app.REMOTING_API).

    See http://www.sencha.com/products/js/direct.php for further infos.

    The optional ``expose_exceptions`` argument controls the output of
    an ExtDirect call - if ``True``, the router will provide additional
    information about exceptions.
    """

    implements(IExtdirect)

    def __init__(self, app, api_path, router_path, namespace='Ext.app', descriptor='Ext.app.REMOTING_API', expose_exceptions=True):
        self.app = app
        self.api_path = api_path
        self.router_path = router_path
        self.namespace = namespace
        self.descriptor = descriptor
        self.expose_exceptions = expose_exceptions

        self.actions = {}

        self.scanned = False

    def add_action(self, action_name, **settings):
        """
        Registers an action.

        ``action_name``: Action name

        Possible values of `settings``:

        ``method_name``: Method name
        ``callback``: The callback to execute upon client request
        ``numargs``: Number of arguments passed to the wrapped callable
        ``scope``: class/module/exec/etc., see venusian.advice.getFrameInfo()
        ``accept_files``: If true, this action will be declared as formHandler in API
        ``instance_name``: In case we're dealing with an instance method ``instance_name``
            is the name of the traversable object in BFG traversal graph
        ``permission``: The permission needed to execute the wrapped callable
        ``request_as_last_param``: If true, the wrapped callable will receive a request object
            as last argument

        """
        if action_name not in self.actions:
            self.actions[action_name] = {}
        self.actions[action_name][_mk_cb_key(action_name, settings['method_name'])] = settings

    def get_actions(self):
        """ Builds and returns a dict of actions to be used in ExtDirect API """
        ret = {}
        for k, v in self.actions.items():
            itms = []
            for settings in v.values():
                d = dict(
                    len = settings['numargs'],
                    name = settings['method_name']
                )
                if settings['accepts_files']:
                    d['formHandler'] = True
                itms.append(d)
            ret[k] = itms
        return ret

    def get_method(self, action, method):
        """ Returns a method's settings """
        if action not in self.actions:
            raise KeyError("Invalid action: " + action)
        key = _mk_cb_key(action, method)
        if not key in self.actions[action]:
            raise KeyError("No such method in '%s': '%s':" % (action, method))
        return self.actions[action][key]

    def dump_api(self, request):
        """ Dumps all known remote methods """
        ret = ["Ext.ns('%s'); %s = " % (self.namespace, self.descriptor)]
        apidict = dict(
            url = request.application_url + '/' + self.router_path,
            type = 'remoting',
            namespace = self.namespace,
            actions = self.get_actions()
        )
        ret.append(json.dumps(apidict))
        ret.append(";")
        return "".join(ret)

    def _do_route(self, action_name, method_name, params, transaction_id, request):
        """ Performs routing, i.e. calls decorated methods/functions """
        settings = self.get_method(action_name, method_name)
        permission = settings.get('permission', None)
        ret = {
            "type": "rpc",
            "tid": transaction_id,
            "action": action_name,
            "method": method_name,
            "result": None
        }
        req_as_last = settings.get('request_as_last_param', False)
        if params is None:
            params = list()
        if req_as_last:
            params.append(request)
        try:
            callback = settings['callback']
            if settings['scope'] == 'class':
                # FIXME this relies heavily on BFGs model traversal
                instance = request.root[settings['instance_name']]
                if permission is not None and not has_permission(permission, instance, request):
                    raise Exception("Access denied")
                params.insert(0, instance)
                if callback is None or not callable(callback):
                    raise Exception("Invalid method '%s' for action '%s'" % (method_name, action_name,))
            ret["result"] = callback(*params)
        except Exception, e:
            ret["type"] = "exception"
            if self.expose_exceptions:
                f = StringIO()
                traceback.print_exc(file=f)
                ret["result"] = {'error': True, 'message': str(e), 'exception_class': str(Exception), 'stacktrace': f.getvalue()}
            else:
                ret["result"] = {'error': True, 'message': 'Error executing %s.%s' % (action_name, method_name,)}
        return ret

    def route(self, request):
        is_form_data = False
        if is_form_submit(request):
            is_form_data = True
            params = parse_extdirect_form_submit(request)
        else:
            params = parse_extdirect_request(request)
        ret = []
        for (act, meth, params, tid) in params:
            ret.append(self._do_route(act, meth, params, tid, request))
        if not is_form_data:
            if len(ret) == 1:
                ret = ret[0]
            return (json.dumps(ret, cls=JsonReprEncoder), False)
        ret = ret[0] # form data cannot be batched
        s = json.dumps(ret, cls=JsonReprEncoder).replace('&quot;', '\\&quot;')
        return (FORM_SUBMIT_RESPONSE_TPL % (s,), True)


class extdirect_method(object):
    """ Enables direct extjs access to python methods through json/form submit """

    def __init__(self, action=None, method_name=None, permission=None, accepts_files=False, request_as_last_param=False):
        self._settings = dict(
            action = action,
            method_name = method_name,
            permission = permission,
            accepts_files = accepts_files,
            request_as_last_param = request_as_last_param,
        )

    @property
    def settings(self):
        return self._settings.copy()

    def register(self, scanner, name, ob):
        info = self.info
        settings = self.settings
        wrapped = self.wrapped

        (args, varargs, varkw, defaults) = inspect.getargspec(wrapped)
        numargs = len(args)

        if numargs and settings['request_as_last_param']:
            numargs -= 1

        settings['numargs'] = numargs

        scope = info.scope
        settings['scope'] = scope
        if scope == 'class':
            settings['numargs'] -= 1
            if '__acl__' in info.locals:
                settings['acl'] = info.locals['__acl__']
            if settings.get('method_name') is None:
                settings['method_name'] = wrapped.__name__
            settings['instance_name'] = None
            if '__name__' in info.locals:
                settings['instance_name'] = info.locals['__name__']
        elif scope == 'module':
            if settings.get('method_name') is None:
                settings['method_name'] = wrapped.func_name

        action = settings.get("action")
        if action is not None:
            name = action
        elif scope == 'class':
            extdirect_settings = getattr(ob, '__extdirect_settings__', None)
            if extdirect_settings is not None:
                if 'default_action_name' in extdirect_settings:
                    name = extdirect_settings['default_action_name']
                if settings.get("permission") is None\
                        and 'default_permission' in extdirect_settings:
                    settings['permission'] = extdirect_settings['default_permission']
        if 'action' in settings:
            del settings['action']

        extdirect = scanner.config.registry.getUtility(IExtdirect)
        extdirect.add_action(name, callback=wrapped, **settings)

    def __call__(self, wrapped):
        self.info = venusian.attach(wrapped,
                                    self.register,
                                    category='extdirect')
        self.wrapped = wrapped
        return wrapped


def is_form_submit(request):
    """ Checks if a request contains extdirect form submit """
    p = request.params
    keys = list(FORM_DATA_KEYS)
    for key in p.keys():
        if key in keys:
            keys.remove(key)
    return len(keys) == 0

def parse_extdirect_form_submit(request):
    """
        Extracts extdirect remoting parameters from request
        which are provided by a form submission
    """
    p = request.params
    if not p:
        raise Exception("Could not parse form submit")
    action = p.get('extAction')
    method = p.get('extMethod')
    tid = p.get('extTID')
    data = {}
    for key in p.keys():
        if key in FORM_DATA_KEYS:
            continue
        data[key] = p.get(key)
    return [(action, method, [data], tid)]

def parse_extdirect_request(request):
    """
        Extracts extdirect remoting parameters from request
        which are provided by an AJAX request
    """
    body = request.body
    decoded_body = json.loads(body)
    ret = []
    if type(decoded_body) is not list:
        decoded_body = [decoded_body]
    for p in decoded_body:
        action = p['action']
        method = p['method']
        data = p['data']
        tid = p['tid']
        ret.append((action, method, data, tid))
    return ret


def api_view(request):
    """ Renders the API """
    util = request.registry.getUtility(IExtdirect)
    body = util.dump_api(request)
    return Response(body, content_type='text/javascript', charset='UTF-8')

def router_view(request):
    """ Renders the result of a ExtDirect call """
    util = request.registry.getUtility(IExtdirect)
    (body, is_form_data) = util.route(request)
    ctype = 'application/json'
    if is_form_data:
        ctype = 'text/html'
    return Response(body, content_type=ctype, charset='UTF-8')
